/**
 * \file basecell.h
 *
 * \brief This is, more or less, the "library" of models.              
 *
 */

#ifndef BASECELL_H
#define BASECELL_H

#include "macros.h"
#include "types.h"
#include <cstring>
#include <string>




/** \brief Enumeration of basecells of different 1D and 2D models. */
enum MODEL {
  UNDEF,
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  // 1-DIMENSIONAL MODELS: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  // ID                 SPINS
  HB1,       /**< (1D) 1/2 - 1/2 - 1/2 - 1/2 */
  HB2,       /**< (1D)  1 -   1  -  1  -  1  */
  HB3,       /**< (1D) 3/2 - 3/2 - 3/2 - 3/2 */
  MA,        /**< (1D) 1/2 - 1/2 -  1  -  1  */
  MB,        /**< (1D) 1/2 - 1/2 - 3/2 - 3/2 */
  MC,        /**< (1D)  1  -  1  - 3/2 - 3/2 */
  ferriMA,   /**< (1D) 1/2 -  1              */
  ferriMB,   /**< (1D) 1/2 - 3/2             */
  ferriMC,   /**< (1D)  1  - 3/2             */
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  // 1-D RANDOM DOPING MODELS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  HB1dop2,  /**< (1D) mix of spin 1/2 and 1       */
  MAdop1,   /**< (1D) replace in MA S=1 by S=1/2  */
  MAdop2,   /**< (1D) replace in MA S=1/2 by S=1  */
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  // 2-DIMENSIONAL MODELS              %%%%%%%%%%%%%%%%%%%%
  // i refers to the row index, i.e. y %%%%%%%%%%%%%%%%%%%%
  // j refers to the col index, i.e. x %%%%%%%%%%%%%%%%%%%%
  // (see below in Class Basecell)     %%%%%%%%%%%%%%%%%%%%
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  // -----------------------------------------------
  // J-J' MODELS
  // -----------------------------------------------
  //   vertical bonds y[i,j] are J (set to unity!!!)
  // horizontal bonds x[i,j] are J' if (i+j)%2 == 0
  //                         and J  if (i+j)%2 == 1
  // -----------------------------------------------
  JJpHB1, /**< (2D J-J') S=1/2 */
  JJpHB2, /**< (2D J-J') S= 1  */
  JJpHB3, /**< (2D J-J') S=3/2 */
  // -----------------------------------------------

  // -----------------------------------------------
  // PLAQUETTE MODELS
  // -----------------------------------------------
  //   vertical bonds y[i,j] are J  if i%2 == 1
  //                         and J' if i%2 == 0
  // horizontal bonds x[i,j] are J  if j%2 == 1
  //                         and J' if j%2 == 0
  // -----------------------------------------------
  plaqHB1, /**< (2D plaquette) S=1/2 */
  plaqHB2, /**< (2D plaquette) S= 1  */
  plaqHB3, /**< (2D plaquette) S=3/2 */
  // -----------------------------------------------

  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  // 1-D MODEL BEND ON A 2-D LATTICE   %%%%%%%%%%%%%%%%%%%%
  // for test purposes only            %%%%%%%%%%%%%%%%%%%%
  // "chlt" means chain on lattice     %%%%%%%%%%%%%%%%%%%%
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  chltHB1, /**< (1D chain on a 2D lattice) S=1/2 */
  chltHB2, /**< (1D chain on a 2D lattice) S= 1  */
  chltHB3, /**< (1D chain on a 2D lattice) S=3/2 */
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  UNDEF2
};


/******************************************************//**
* \brief A Basecell contains information about the spinsize S,
*  and the couplings Jx,Jy,Jz between the spins.
*  
*  The size of the (2D) Basecell is Lx times Ly.
*  
*  Spinsizes are represented by their number of subspins, i.e.
*       S = 1/2 by 1,
*       S = 1   by 2,
*       S = 3/2 by 3, ...
*
*  Spinsizes are stored in a matrix (mat_i) called "S" 
*  and bond strengths are stored in three matrices (mat_d's)
*  called xJx, xJy, xJz and yJx, yJy, yJz, i.e.
*  - Lx = xJx[0].size(),
*  - Ly = xJx.size(), or
*  - Lx = S[0].size(),
*  - Ly = S.size(),
*
*  (all mat_d's as well as "S" are generated with 
*  same dimensions), such that:
*  
*  - S[i][j]    ... spinsize at posX=j (col), posY=i (row)
*  - f[i][j]    ... is spin at posX=j (col), posY=i (row) staggered?
*  - xJn[i][j]  ... component n of the coupling between spin (i,j) and spin (i,j+1)
*  - yJn[i][j]  ... component n of the coupling between spin (i,j) and spin (i+1,j)
*  - x/yA[i][j] ... is bond x/y(i,j) an alpha-bond?
*    
*  The construction of the Basecell is determined by
*  the input parameter "model".
**********************************************************/

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//    *********************************************
//    Illustration of the description of a Basecell
//    (i,j) -> site
//    [i,j] -> bond
//    *********************************************
//    SITES (S):
//      .         .         .         .
//      .         .         .         .
//      .         .         .         .
//    (2,0) ... (2,1) ... (2,2) ... (2,3) ...
//      .         .         .         .
//      .         .         .         .
//      .         .         .         .
//    (1,0) ... (1,1) ... (1,2) ... (1,3) ...
//      .         .         .         .
//      .         .         .         .
//      .         .         .         .
//    (0,0) ... (0,1) ... (0,2) ... (0,3) ...
// 
//    *********************************************
//    HORIZONTAL BONDS (xJn):
//      .         .         .         .
//      .         .         .         .
//      .  x[2,1] .  x[2,2] .  x[2,3] .  x[2,4]
//      O   ...   O   ...   O   ...   O   ...
//      .         .         .         .
//      .         .         .         .
//      .  x[1,1] .  x[1,2] .  x[1,3] .  x[1,4]
//      O   ...   O   ...   O   ...   O   ...
//      .         .         .         .
//      .         .         .         .
//      .  x[0,1] .  x[0,2] .  x[0,3] .  x[0,4]
//      O   ...   O   ...   O   ...   O   ...
// 
//    *********************************************
//    VERTICAL BONDS (yJn):
//      .         .         .         .
//      .y[2,1]   .y[2,2]   .y[2,3]   .y[2,4]
//      .         .         .         .
//      O   ...   O   ...   O   ...   O   ...
//      .         .         .         .
//      .y[1,1]   .y[1,2]   .y[1,3]   .y[1,4]
//      .         .         .         .
//      O   ...   O   ...   O   ...   O   ...
//      .         .         .         .
//      .y[0,1]   .y[0,2]   .y[0,3]   .y[0,4]
//      .         .         .         .
//      O   ...   O   ...   O   ...   O   ...
// 
//    *********************************************
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class Basecell {

 public:
  MODEL model;
  int Lx,Ly;
  mat_i S;
  mat_b f;
  mat_d xJx, xJy, xJz, yJx, yJy, yJz;
  mat_b xA,yA; // tells about alpha bonds
  double alpha,Delta;
  bool modified;
  
  Basecell():Lx(0),Ly(0),alpha((double)1.),Delta((double)1.){};
  ~Basecell(){};

  void setSize(int x, int y){
    Lx = x;
    Ly = y;
  };

  void duplicate_in_x() {
    for (int col=Lx;col<2*Lx;col++) {
      S  [0][col] = S  [0][col-Lx];
      f  [0][col] = f  [0][col-Lx];
      xJx[0][col] = xJx[0][col-Lx];
      xJy[0][col] = xJy[0][col-Lx];
      xJz[0][col] = xJz[0][col-Lx];
      xA [0][col] = xA [0][col-Lx];
      yJx[0][col] = yJx[0][col-Lx];
      yJy[0][col] = yJy[0][col-Lx];
      yJz[0][col] = yJz[0][col-Lx];
      yA [0][col] = yA [0][col-Lx];
    }
    Lx *= 2;
  };


  int get_Lx(std::string mod) {return Lx;};
  // alpha = J'/J, J is set to unity!
  // Ly ... number of "rows"           
  // Lx ... number of "columns"        
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  // models with two coupling inputs, alpha & Delta
  // bool==true -> modified: Delta only on non-alpha bonds
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  void aD_couplings_1D(double J, double d1) {
    for (int col=0;col<Lx;col++) {
      f[0][col] = ((col%2)==1);
      xJx[0][col] = ((col%2)==0)?J:J*alpha;
      xJy[0][col] = ((col%2)==0)?J:J*alpha;
      xJz[0][col] = ((col%2)==0)?J*Delta:J*d1;
      xA [0][col] = ((col%2)!=0);
      yJx[0][col] = 0.;
      yJy[0][col] = 0.;
      yJz[0][col] = 0.;
      yA [0][col] = false;
    }
  };

  bool generate(MODEL mod,double a,double D,bool modi) {
    model = mod;
    alpha = a;
    Delta = D;
    modified = modi;
    bool result = true;
    double J = (double)1.;
    double d1 = (modified)?alpha:alpha*Delta;
    
    int s1=3,s2=3;
    switch(model) {
      //----------------------------------------------------
      // chain models (1D) ---------------------------------
      //----------------------------------------------------
    case HB1: s1--;
    case HB2: s1--;
    case HB3: 
      Ly = 1; Lx = 2;
      for (int col=0;col<Lx;col++) S[0][col] = s1;
      aD_couplings_1D(J,d1);
      break;
    
    case MA: s2--;
    case MB: s1--;
    case MC: s1--;
      Ly = 1; Lx = 4;
      for (int col=0;col<Lx;col++) S[0][col] = (col<2)?s1:s2;
      aD_couplings_1D(J,d1);
      break;
      
    case ferriMA: s2--;
    case ferriMB: s1--;
    case ferriMC: s1--;
      Ly = 1; Lx = 2;
      S[0][0] = s1;
      S[0][1] = s2;
      aD_couplings_1D(J,d1);
      break;
      
      //----------------------------------------------------
      // array models (2D) ---------------------------------
      //----------------------------------------------------
    case JJpHB1: s1--;
    case JJpHB2: s1--;
    case JJpHB3: 
      if (alpha == J) {Lx = 1;Ly = 1;}
      else {Lx = 2;Ly = 2;}
      for (int row=0;row<Lx;row++) {
	for (int col=0;col<Lx;col++) {
	  S[row][col] = s1;
	  f[row][col] = ((row+col)%2==1);
	  xJx[row][col] = (((row+col)%2)==1)?J:J*alpha;
	  xJy[row][col] = (((row+col)%2)==1)?J:J*alpha;
	  xJz[row][col] = (((row+col)%2)==1)?J*Delta:J*d1;
	  xA [row][col] = (((row+col)%2)!=1);
	  yJx[row][col] = J;
	  yJy[row][col] = J;
	  yJz[row][col] = J*Delta;
	  yA [row][col] = false;
	}
      }
      break;
      
    case plaqHB1: s1--;
    case plaqHB2: s1--;
    case plaqHB3:
      Lx = 2; Ly = 2;
      for (int row=0;row<Ly;row++) {
	for (int col=0;col<Lx;col++) {
	  S[row][col] = s1;
	  f[row][col] = (((row+col)%2)==1);
	  xJx[row][col] = ((col%2)==1)?J:J*alpha;
	  xJy[row][col] = ((col%2)==1)?J:J*alpha;
	  xJz[row][col] = ((col%2)==1)?J*Delta:J*d1;
	  xA [row][col] = ((col%2)!=1);
	  yJx[row][col] = ((row%2)==1)?J:J*alpha;
	  yJy[row][col] = ((row%2)==1)?J:J*alpha;
	  yJz[row][col] = ((row%2)==1)?J*Delta:J*d1;
	  yA [row][col] = ((row%2)!=1);
	}
      }
      break;
      
      //----------------------------------------------------
      // chain on a lattice toy models ---------------------
      //----------------------------------------------------
    case chltHB1: s1--;
    case chltHB2: s1--;
    case chltHB3: 
      if (Lx == 0 || Ly == 0) {
	showe("For your choice of model you must first set the size (Lx,Ly) of the basecell");
	result = false;
      }
      else if ((Ly%2) == 1) {
	showe("For your choice of model you must set the number of rows Ly to an EVEN value");
	result = false;
      }
      else {
	for (int row=0;row<Ly;row++)
	  for (int col=0;col<Lx;col++) {
	    S[row][col] = s1;
	    f[row][col] = (((row+col)%2)==1);
	    xA[row][col] = false;
	    yA[row][col] = false;
	  }
	
	for (int row=0;row<Ly;row++) {
	  for (int col=0;col<Lx-1;col++) {
	    xJx[row][col] = J;
	    xJy[row][col] = J;
	    xJz[row][col] = J;
	  }
	  xJx[row][Lx-1] = (double)0.;
	  xJy[row][Lx-1] = (double)0.;
	  xJz[row][Lx-1] = (double)0.;
	}
	for (int row=0;row<Ly;row++) {
	  for (int col=0;col<Lx-1;col++) {
	    yJx[row][col] = (double)0.;
	    yJy[row][col] = (double)0.;
	    yJz[row][col] = (double)0.;
	  }
	}
	for (int row=0;row<Ly;row++) {
	  if ((row%2) == 0) {
	    yJx[row][Lx-1] = J;
	    yJy[row][Lx-1] = J;
	    yJz[row][Lx-1] = J;
	  }
	  else {
	    yJx[row][0] = J;
	    yJy[row][0] = J;
	    yJz[row][0] = J;
	  }
	}      
      }
      break;
      
    case HB1dop2:
    case MAdop1:
    case MAdop2:  showe(translate(model) << " not defined, use generate_random_doping instead!");
    case UNDEF: break;
    }
    
    return result;
  };
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  // models with one coupling input, alpha only
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  bool generate(MODEL mod, double a) {return generate(mod,a,(double)1.,false);};
  // models with two coupling inputs, alpha & Delta
  bool generate(MODEL mod, double a, double D) {return generate(mod,a,D,false);};

  bool generate_two_spins(MODEL mod,double D,int comb) {
    model = mod;
    alpha = (double)1.;
    Delta = D;
    bool result = true;
    double J = (double)1.;
    
    int s1=3,s2=3;
    switch(model) {
      //--------------------------------------------------
      // chain models (1D) -------------------------------
      //--------------------------------------------------
    case HB1: s1--;
    case HB2: s1--;
    case HB3: Ly=1; Lx=2;
      for (int col=0;col<Lx;col++) {
	S[0][col] = s1;
	f[0][col] = ((col%2)==1);
	xJx[0][col] = ((col%2)==0)?J:0.;
	xJy[0][col] = ((col%2)==0)?J:0.;
	xJz[0][col] = ((col%2)==0)?J*Delta:0.;
	xA [0][col] = ((col%2)!=0);
	yJx[0][col] = 0.;
	yJy[0][col] = 0.;
	yJz[0][col] = 0.;
	yA [0][col] = false;
      }
      break;
      
    case MA: s2--;
    case MB: s1--;
    case MC: s1--; Ly=1; Lx=2;
      switch(comb) {
      case 1: S[0][0] = s1; S[0][1] = s2; break;
      case 2: S[0][0] = s2; S[0][1] = s2; break;
      default: S[0][0] = s1; S[0][1] = s1;
      }
      for (int col=0;col<Lx;col++) {
	f[0][col] = ((col%2)==1);
	xJx[0][col] = ((col%2)==0)?J:0.;
	xJy[0][col] = ((col%2)==0)?J:0.;
	xJz[0][col] = ((col%2)==0)?J*Delta:0.;
	xA [0][col] = ((col%2)!=0);
	yJx[0][col] = 0.;
	yJy[0][col] = 0.;
	yJz[0][col] = 0.;
	yA [0][col] = false;
      }
      break;
      
    case UNDEF:
    default:    result=false; break;
    }
    
    return result;
  };
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  void sow(uint s) {srand(s);};
  bool generate_random_doping(MODEL mod,double a,int lx,double p) {
    model = mod;
    alpha = a;
    Delta = 1.0;
    modified = false;
    bool result = true;
    double J = (double)1.;
    double d1 = (modified)?alpha:alpha*Delta;
    double r;
    bool same;
    
    int s1=3,s2=3;
    switch(model) {
      //----------------------------------------------------
      // chain models (1D) ---------------------------------
      //----------------------------------------------------
    case MA: s2--;
    case MB: s1--;
    case MC: s1--; Lx=lx; Ly=1;
      for (int col=0;col<Lx;col++) {
	r = (double)rand()/(double)RAND_MAX;
	S[0][col] = (r<p)?s2:s1;
      }
      for (int col=0;col<Lx;col++) {
	f[0][col] = ((col%2)==1);
	same = (S[0][col] == S[0][(col+1)%Lx]);
	xJx[0][col] = (same)?J:J*alpha;
	xJy[0][col] = (same)?J:J*alpha;
	xJz[0][col] = (same)?J*Delta:J*d1;
	xA [0][col] = (same)?false:true;
	yJx[0][col] = 0.;
	yJy[0][col] = 0.;
	yJz[0][col] = 0.;
	yA [0][col] = false;
      }
      break;

    case MAdop1: s1=1; s2=2; Lx=lx; Ly=1;
      for (int col=0;col<Lx/4;col++) {
	S[0][4*col] = s1;
	S[0][4*col+1] = s1;
	r = (double)rand()/(double)RAND_MAX;
	S[0][4*col+2] = (r<p)?s1:s2;
	r = (double)rand()/(double)RAND_MAX;
	S[0][4*col+3] = (r<p)?s1:s2;
      }
      for (int col=0;col<Lx;col++) {
	f[0][col] = ((col%2)==1);
	same = (S[0][col] == S[0][(col+1)%Lx]);
	xJx[0][col] = (same)?J:J*alpha;
	xJy[0][col] = (same)?J:J*alpha;
	xJz[0][col] = (same)?J*Delta:J*d1;
	xA [0][col] = (same)?false:true;
	yJx[0][col] = 0.;
	yJy[0][col] = 0.;
	yJz[0][col] = 0.;
	yA [0][col] = false;
      }
      break;

    case MAdop2: s1=1; s2=2; Lx=lx; Ly=1;
      for (int col=0;col<Lx/4;col++) {
	r = (double)rand()/(double)RAND_MAX;
	S[0][4*col] = (r<p)?s2:s1;
	r = (double)rand()/(double)RAND_MAX;
	S[0][4*col+1] = (r<p)?s2:s1;
	S[0][4*col+2] = s2;
	S[0][4*col+3] = s2;
      }
      for (int col=0;col<Lx;col++) {
	f[0][col] = ((col%2)==1);
	same = (S[0][col] == S[0][(col+1)%Lx]);
	xJx[0][col] = (same)?J:J*alpha;
	xJy[0][col] = (same)?J:J*alpha;
	xJz[0][col] = (same)?J*Delta:J*d1;
	xA [0][col] = (same)?false:true;
	yJx[0][col] = 0.;
	yJy[0][col] = 0.;
	yJz[0][col] = 0.;
	yA [0][col] = false;
      }
      break;
      
    case UNDEF: 
    default: 
      showe("No random model defined for " << translate(model));
      result = false;
      break;
    }
    return result;
  };
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  // TRANSLATE BETWEEN "MODEL" AND "STRING" %%%%%%%%%%%%%%%
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  std::string translate(MODEL M) {
    if      (M == HB1)     return("HB1");
    else if (M == HB2)     return("HB2");
    else if (M == HB3)     return("HB3");
    else if (M == MA)      return("MA");
    else if (M == MB)      return("MB");
    else if (M == MC)      return("MC");
    else if (M == ferriMA) return("ferriMA");
    else if (M == ferriMB) return("ferriMB");
    else if (M == ferriMC) return("ferriMC");
    else if (M == JJpHB1)  return("JJpHB1");
    else if (M == JJpHB2)  return("JJpHB2");
    else if (M == JJpHB3)  return("JJpHB3");
    else if (M == plaqHB1) return("plaqHB1");
    else if (M == plaqHB2) return("plaqHB2");
    else if (M == plaqHB3) return("plaqHB3");
    else if (M == chltHB1) return("chltHB1");
    else if (M == chltHB2) return("chltHB2");
    else if (M == chltHB3) return("chltHB3");
    else if (M == HB1dop2) return("HB1dop2");
    else if (M == MAdop1)  return("MAdop1");
    else if (M == MAdop2)  return("MAdop2");
    else                   return("undefined");
  };

  MODEL translate(std::string m) {
    if      (m.compare("HB1")     == 0) return HB1;
    else if (m.compare("HB2")     == 0) return HB2;
    else if (m.compare("HB3")     == 0) return HB3;
    else if (m.compare("MA")      == 0) return MA;
    else if (m.compare("MB")      == 0) return MB;
    else if (m.compare("MC")      == 0) return MC;
    else if (m.compare("ferriMA") == 0) return ferriMA;
    else if (m.compare("ferriMB") == 0) return ferriMB;
    else if (m.compare("ferriMC") == 0) return ferriMC;
    else if (m.compare("JJpHB1")  == 0) return JJpHB1;
    else if (m.compare("JJpHB2")  == 0) return JJpHB2;
    else if (m.compare("JJpHB3")  == 0) return JJpHB3;
    else if (m.compare("plaqHB1") == 0) return plaqHB1;
    else if (m.compare("plaqHB2") == 0) return plaqHB2;
    else if (m.compare("plaqHB3") == 0) return plaqHB3;
    else if (m.compare("chltHB1") == 0) return chltHB1;
    else if (m.compare("chltHB2") == 0) return chltHB2;
    else if (m.compare("chltHB3") == 0) return chltHB3;
    else if (m.compare("HB1dop2") == 0) return HB1dop2;
    else if (m.compare("MAdop1")  == 0) return MAdop1;
    else if (m.compare("MAdop2")  == 0) return MAdop2;
    else                                return UNDEF;
  };
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  // DISPLAY BASECELL ON TERMINAL SCREEN %%%%%%%%%%%%%%%%%%
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  void display(){
    std::string strModel;
    strModel = translate(model);
    
    showlog("Model is " << strModel);
    showlog("BASECELL:");
    
    for (int row=Ly-1;row>=0;row--) {
      // VERTICAL BONDS
      for (int col=0;col<Lx;col++) 
	showlogl("|x: " << std::setw(4) << yJx[row][col] << "|" << yA[row][col] << "            ");
      showlog("");
      for (int col=0;col<Lx;col++)
	showlogl("|y: " << std::setw(4) << yJy[row][col] << "|             ");
      showlog("");
      for (int col=0;col<Lx;col++) 
	showlogl("|z: " << std::setw(4) << yJz[row][col] << "|             ");
      showlog("");
      for (int col=0;col<Lx;col++) showlogl("    |                 ");
      showlog("");
      showlogl("   ");
      
      
      // SPINS AND HORIZONTAL BONDS
      for (int col=0;col<Lx;col++) {
	showlogl(" |      ");
	showlogl("|x: " << std::setw(4) << xJx[row][col] << "|" << xA[row][col] << "    ");
      }
      showlog("");
      showlogl("-- ");
      for (int col=0;col<Lx;col++) {
	showlogl("|" << S[row][col] << "|" << f[row][col] << "--- ");
	showlogl("|y: " << std::setw(4) << xJy[row][col] << "| --- ");
      }
      showlog("");
      showlogl("   ");
      for (int col=0;col<Lx;col++) {
	showlogl(" |      ");
	showlogl("|z: " << std::setw(4) << xJz[row][col] << "|     ");
      }
      showlog("");
      for (int col=0;col<Lx;col++) showlogl("    |                 ");
      showlog("");
    }
  };
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#endif // BASECELL_H ///:~
