/**
 * \file measurement_producer.h
 *
 * \brief Help to organize measurements.
 *
 */

#ifndef MEASUREMENT_PRODUCER_H
#define MEASUREMENT_PRODUCER_H

#include "measurements.h"




/** \brief Enumeration of estimators. */
enum ESTIMATOR {
  UNDEFINED, // undefined
  ENERGY,		/**< "E" ... total internal energy */
  aENERGY,		/**< "aE" ... total internal energy on alternated bonds */
  dENERGY,		/**< "dE" ... total diagonal energy */
  nOPERATORS,		/**< "nO" ... total number of operators in a configuration */
  dnOPERATORS,		/**< "dnO" ... total number of operators in a decorated configuration */
  daENERGY,		/**< "daE" ... diagonal energy on alternated bonds */
  naOPERATORS,		/**< "naO" ... number of operators on alternated bonds in a configuration */
  dnaOPERATORS,		/**< "naO" ... number of operators on alternated bonds in a decorated configuration */
  M,			/**< "m" ... magnetization */
  slSTM,		/**< "sl_stm" ... slice estimator of staggered magnetization */
  inSTM,		/**< "in_stm" ... integrated estimator of staggered magnetization */
  inSTM0,		/**< "in_stm0" ... integrated estimator of absolute value of staggered magnetization */
  inSTM2,		/**< "in_stm2" ... integrated estimator of squared staggered magnetization */
  inSTM4,		/**< "in_stm4" */
  slTWIST,		/**< "sl_tw" ... slice estimator of twist parameter */ 
  inTWIST,		/**< "in_tw" ... integrated estimator of twist parameter */ 
  inTWIST0,		/**< "in_tw0" ... integrated estimator of absolute value of twist parameter */ 
  inTWIST2,		/**< "in_tw2" ... integrated estimator of squared twist parameter */ 
  inTWIST4,		/**< "in_tw4" */ 
  PAIR,			/**< "pair" ... pair correlation */ 
  avPAIR,		/**< "av_pair" ... averaged pair correlation */ 
  //------------------------
  // by default <SZSZ>:
  impSTSUS,		/**< "imp_stsus" ... longitudinal staggered susceptibility */ 
  impTWOPIb,		/**< "imp_2PIb" ... Fourier component of longitudinal dynamic structure factor */ 
  impTWOPIL,		/**< "imp_2PIL" ... Fourier component of longitudinal dynamic structure factor */ 
  impFOURPIb,		/**< "imp_4PIb" ... Fourier component of longitudinal dynamic structure factor */ 
  impFOURPIL,		/**< "imp_4PIL" ... Fourier component of longitudinal dynamic structure factor */ 
  impSIXPIb,		/**< "imp_6PIb" ... Fourier component of longitudinal dynamic structure factor */ 
  impSIXPIL,		/**< "imp_6PIL" ... Fourier component of longitudinal dynamic structure factor */ 
  //------------------------
  // <S+S->
  impSTSUSPM,		/**< "imp_stsuspm" ... transversal staggered susceptibility */ 
  impTWOPIbPM,		/**< "imp_2PIbpm" ... Fourier component of transversal dynamic structure factor */ 
  impTWOPILPM,		/**< "imp_2PILpm" ... Fourier component of transversal dynamic structure factor */ 
  impFOURPIbPM,		/**< "imp_4PIbpm" ... Fourier component of transversal dynamic structure factor */ 
  impFOURPILPM,		/**< "imp_4PILpm" ... Fourier component of transversal dynamic structure factor */ 
  impSIXPIbPM,		/**< "imp_6PIbpm" ... Fourier component of transversal dynamic structure factor */ 
  impSIXPILPM,		/**< "imp_6PILpm" ... Fourier component of transversal dynamic structure factor */ 
  //------------------------
  impTWN2,		/**< "imp_twN2" ... squared temporal winding number */ 
  impSWN2		/**< "imp_swN2" ... squared spatial winding number */
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// MEASUREMENT PRODUCER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class MeasurementProducer {
  Spinsystem* SA;
  int energy;
  int in_tw;
  int imp_stsus;
  int imp_stsuspm;
  std::string s;
  ESTIMATOR e;

 public:
  MeasurementProducer(Spinsystem& sa):SA(&sa),energy(-1),in_tw(-1),imp_stsus(-1),imp_stsuspm(-1){};
  ~MeasurementProducer(){};

  /**
   * \brief An attempt to avoid unnecessary double measurements.
   *
   * Some measurements can "take" other measurements in order to avoid
   * double calculations.
   *
   */
  void do_work(std::vector<Measurement*>& est, std::vector<double>& norm) {
    e = translate(s);
    if (e != UNDEFINED) {
      if (e==ENERGY)     energy       = est.size();
      if (e==inTWIST)    in_tw        = est.size();
      if (e==impSTSUS)   imp_stsus    = est.size();
      if (e==impSTSUSPM) imp_stsuspm  = est.size();
      est.push_back(get_measurement(e));
      norm.push_back(get_norm(e));
      
      if (e==dENERGY     && energy != -1) est[energy]->take_dE (est.back());
      if (e==nOPERATORS  && energy != -1) est[energy]->take_nO (est.back());
      if (e==daENERGY    && energy != -1) est[energy]->take_daE(est.back());
      if (e==naOPERATORS && energy != -1) est[energy]->take_naO(est.back());
      
      if (e==inTWIST0 || e==inTWIST2 || e==inTWIST4) 
	if (in_tw != -1) est[in_tw]->take(est.back());
      
      if (e==impTWOPIb||e==impTWOPIL||e==impFOURPIb||e==impFOURPIL||e==impSIXPIb||e==impSIXPIL)
	if (imp_stsus != -1) est[imp_stsus]->take(est.back());
      
      if (e==impTWOPIbPM||e==impTWOPILPM||e==impFOURPIbPM||e==impFOURPILPM||e==impSIXPIbPM||e==impSIXPILPM)
	if (imp_stsuspm != -1) est[imp_stsuspm]->take(est.back());
    }
  };

  void read_measurements(std::string name, std::vector<Measurement*>& est, std::vector<double>& norm){
    read_measurements(name,est,norm,"");
  };
  void read_measurements(std::string name, std::vector<Measurement*>& est, std::vector<double>& norm, std::string but_not) {
    std::ifstream in(name.c_str());
    if (in.is_open()) {
      in >> s;
      if (!s.empty() && s.compare(but_not)==0) {showe("WARNING!!! " << s << " should not be read in here!!!"); exit(0);}
      if (!in.eof()) {
	if (*s.begin() != '#') do_work(est,norm);
	while(true) {
	  in >> s;
	  if (s.compare(but_not)==0) {showe("WARNING!!! " << s << " should not be read in here!!!"); exit(0);}
	  if (*s.begin() != '#') {
	    if (in.eof()) break;
	    do_work(est,norm);
	  }
	  else if (in.eof()) break;
	}
      }
      in.close();
    }
  };

  Measurement* get_measurement(std::string m) {return get_measurement(translate(m));};
  Measurement* get_measurement(ESTIMATOR Meas) {
    std::string temp;
    switch (Meas) {
    default:   
    case UNDEFINED:   
      showe("Measurement " << Meas);
      temp = translate(Meas);
      showe("Measurement " << temp);
      showe("Measurement " << translate(Meas) << " not defined!!!");
      return NULL;
    case ENERGY:      return new Energy(*SA);
    case aENERGY:     return new AlphaEnergy(*SA);
    case dENERGY:     return new DiagEnergy(*SA);
    case nOPERATORS:  return new NumOp(*SA);
    case dnOPERATORS: return new NumOp(*SA,"dnO");
    case daENERGY:    return new DiagAlphaEnergy(*SA);
    case naOPERATORS: return new NumAlphaOp(*SA);
    case dnaOPERATORS:return new NumAlphaOp(*SA,"dnaO");
    case M:           return new Mag(*SA);
    case slSTM:       return new Stmag(*SA);
    case inSTM:       return new Stmag(*SA,false);
    case inSTM0:      return new AbsStmag(*SA,false);
    case inSTM2:      return new SqStmag(*SA,false);
    case inSTM4:      return new QuStmag(*SA,false);
    case slTWIST:     return new Twist(*SA);
    case inTWIST:     return new Twist(*SA,false);
    case inTWIST0:    return new AbsTwist(*SA,false);
    case inTWIST2:    return new SqTwist(*SA,false);
    case inTWIST4:    return new QuTwist(*SA,false);
    case PAIR:        return new Pair(*SA);
    case avPAIR:      return new AvgPair(*SA);
    case impSTSUS:    return new SqClusterWeight(*SA,"imp_stsus");
    case impTWOPIb:   return new GenSqClusterWeight(*SA,"imp_2PIb",0.,0.,TWOPI/SA->beta);
    case impTWOPIL:   return new GenSqClusterWeight(*SA,"imp_2PIL",TWOPI/(double)SA->Lx,0.,0.);
    case impFOURPIb:  return new GenSqClusterWeight(*SA,"imp_4PIb",0.,0.,2.*TWOPI/SA->beta);
    case impFOURPIL:  return new GenSqClusterWeight(*SA,"imp_4PIL",2.*TWOPI/(double)SA->Lx,0.,0.);
    case impSIXPIb:   return new GenSqClusterWeight(*SA,"imp_6PIb",0.,0.,3.*TWOPI/SA->beta);
    case impSIXPIL:   return new GenSqClusterWeight(*SA,"imp_6PIL",3.*TWOPI/(double)SA->Lx,0.,0.);
    case impSTSUSPM:    return new SqLoopSize(*SA,"imp_stsuspm");
    case impTWOPIbPM:   return new GenSqLoopSize(*SA,"imp_2PIbpm",0.,0.,TWOPI/SA->beta);
    case impTWOPILPM:   return new GenSqLoopSize(*SA,"imp_2PILpm",TWOPI/(double)SA->Lx,0.,0.);
    case impFOURPIbPM:  return new GenSqLoopSize(*SA,"imp_4PIbpm",0.,0.,2.*TWOPI/SA->beta);
    case impFOURPILPM:  return new GenSqLoopSize(*SA,"imp_4PILpm",2.*TWOPI/(double)SA->Lx,0.,0.);
    case impSIXPIbPM:   return new GenSqLoopSize(*SA,"imp_6PIbpm",0.,0.,3.*TWOPI/SA->beta);
    case impSIXPILPM:   return new GenSqLoopSize(*SA,"imp_6PILpm",3.*TWOPI/(double)SA->Lx,0.,0.);
    case impTWN2:       return new TempWindNum2(*SA);
    case impSWN2:       return new SpatialWindNum2(*SA);
    }
  };

  double get_norm(std::string m) {return get_norm(translate(m));};
  double get_norm(ESTIMATOR Meas) {
    double L(SA->Lx),beta(SA->beta),Lh(L/2.);
    switch (Meas) {
    case UNDEFINED:
    case nOPERATORS:
    case naOPERATORS:
    case dnOPERATORS:
    case dnaOPERATORS:
    case slTWIST:
    case PAIR:
    case impTWN2:
      return 1.;

    case M:
    case slSTM:
      return L;

    case inTWIST:
    case inTWIST0:
    case inTWIST2:
    case inTWIST4:
    case ENERGY:
    case aENERGY:
    case dENERGY:
    case daENERGY:
    case impSWN2:
      return beta;

    case inSTM:
    case inSTM0:
    case inSTM2:
    case inSTM4:
      return L*beta;      

    case avPAIR:
      return Lh;

    case impSTSUS:       
    case impTWOPIb:
    case impTWOPIL:
    case impFOURPIb:
    case impFOURPIL:
    case impSIXPIb:
    case impSIXPIL:
    case impSTSUSPM:
    case impTWOPIbPM:
    case impTWOPILPM:
    case impFOURPIbPM:
    case impFOURPILPM:
    case impSIXPIbPM:
    case impSIXPILPM:   
      return 4.*beta*L;

    default: 
      return 1.;
    }
  };


  /** \brief Helps to translate between string and enum handling of estimators. */
  std::string translate(ESTIMATOR Meas) {
    switch (Meas) {
    default:
    case UNDEFINED:   return("undefined");
    case ENERGY:      return("E");
    case aENERGY:     return("aE");
    case dENERGY:     return("dE");
    case nOPERATORS:  return("nO");
    case dnOPERATORS: return("dnO");
    case daENERGY:    return("daE");
    case naOPERATORS: return("naO");
    case dnaOPERATORS:return("dnaO");
    case M:           return("m");
    case slSTM:       return("sl_stm");
    case inSTM:       return("in_stm");
    case inSTM0:      return("in_stm0");
    case inSTM2:      return("in_stm2");
    case inSTM4:      return("in_stm4");
    case slTWIST:     return("sl_tw");
    case inTWIST:     return("in_tw");
    case inTWIST0:    return("in_tw0");
    case inTWIST2:    return("in_tw2");
    case inTWIST4:    return("in_tw4");
    case PAIR:        return("pair");
    case avPAIR:      return("av_pair");
    case impSTSUS:    return("imp_stsus");
    case impTWOPIb:   return("imp_2PIb");
    case impTWOPIL:   return("imp_2PIL");
    case impFOURPIb:  return("imp_4PIb");
    case impFOURPIL:  return("imp_4PIL");
    case impSIXPIb:   return("imp_6PIb");
    case impSIXPIL:   return("imp_6PIL");
    case impSTSUSPM:    return("imp_stsuspm");
    case impTWOPIbPM:   return("imp_2PIbpm");
    case impTWOPILPM:   return("imp_2PILpm");
    case impFOURPIbPM:  return("imp_4PIbpm");
    case impFOURPILPM:  return("imp_4PILpm");
    case impSIXPIbPM:   return("imp_6PIbpm");
    case impSIXPILPM:   return("imp_6PILpm");
    case impTWN2:       return("imp_twN2");
    case impSWN2:       return("imp_swN2");
    }
  };

  /** \brief Helps to translate between string and enum handling of estimators. */
  ESTIMATOR translate(std::string m) {
    if (m.compare("undefined")     == 0) return UNDEFINED;
    else if (m.compare("E")        == 0) return ENERGY;
    else if (m.compare("aE")       == 0) return aENERGY;
    else if (m.compare("dE")       == 0) return dENERGY;
    else if (m.compare("nO")       == 0) return nOPERATORS;
    else if (m.compare("dnO")      == 0) return dnOPERATORS;
    else if (m.compare("daE")      == 0) return daENERGY;
    else if (m.compare("naO")      == 0) return naOPERATORS;
    else if (m.compare("dnaO")     == 0) return dnaOPERATORS;
    else if (m.compare("m")        == 0) return M;
    else if (m.compare("sl_stm")   == 0) return slSTM;
    else if (m.compare("in_stm")   == 0) return inSTM;
    else if (m.compare("in_stm0")  == 0) return inSTM0;
    else if (m.compare("in_stm2")  == 0) return inSTM2;
    else if (m.compare("in_stm4")  == 0) return inSTM4;
    else if (m.compare("sl_tw")    == 0) return slTWIST;
    else if (m.compare("in_tw")    == 0) return inTWIST;
    else if (m.compare("in_tw0")   == 0) return inTWIST0;
    else if (m.compare("in_tw2")   == 0) return inTWIST2;
    else if (m.compare("in_tw4")   == 0) return inTWIST4;
    else if (m.compare("pair")     == 0) return PAIR;
    else if (m.compare("av_pair")  == 0) return avPAIR;
    else if (m.compare("imp_stsus")== 0) return impSTSUS;
    else if (m.compare("imp_2PIb") == 0) return impTWOPIb;
    else if (m.compare("imp_2PIL") == 0) return impTWOPIL;
    else if (m.compare("imp_4PIb") == 0) return impFOURPIb;
    else if (m.compare("imp_4PIL") == 0) return impFOURPIL;
    else if (m.compare("imp_6PIb") == 0) return impSIXPIb;
    else if (m.compare("imp_6PIL") == 0) return impSIXPIL;
    else if (m.compare("imp_stsuspm")== 0) return impSTSUSPM;
    else if (m.compare("imp_2PIbpm") == 0) return impTWOPIbPM;
    else if (m.compare("imp_2PILpm") == 0) return impTWOPILPM;
    else if (m.compare("imp_4PIbpm") == 0) return impFOURPIbPM;
    else if (m.compare("imp_4PILpm") == 0) return impFOURPILPM;
    else if (m.compare("imp_6PIbpm") == 0) return impSIXPIbPM;
    else if (m.compare("imp_6PILpm") == 0) return impSIXPILPM;
    else if (m.compare("imp_twN2")   == 0) return impTWN2;
    else if (m.compare("imp_swN2")   == 0) return impSWN2;
    else return UNDEFINED;
  };

};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




#endif // MEASUREMENT_PRODUCER_H ///:~
