#ifndef MEASUREMENTS_H
#define MEASUREMENTS_H

/* #include "myFiles.h" */
#include "spinsystem.h"
#include <list>
#include <cmath>

typedef std::list<double> dlist;

#define TWOPI 6.283185307179586232



//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// BASECLASS - MEASUREMENT %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Measurement : public dlist {
 public:
  //-------------------------------------------------------
  // VARIABLES --------------------------------------------
  //-------------------------------------------------------
  std::string purename,name,filename;
  bool done;
  //-------------------------------------------------------
  
  //-------------------------------------------------------
  // TOOLS ------------------------------------------------
  //-------------------------------------------------------
  iterator iter;
  Spinsystem* SA;
  double res,RE,IM;
  std::vector<Spin*>::iterator s;
  //-------------------------------------------------------

  //-------------------------------------------------------
  // CONSTRUCTORS & DESTRUCTOR ----------------------------
  //-------------------------------------------------------
  Measurement(Spinsystem& sa):name(""),filename(""),done(false),SA(&sa){};
  Measurement(Spinsystem& sa,std::string n):name(n),filename(""),done(false),SA(&sa){purename=name;};
  ~Measurement(){};
  //-------------------------------------------------------

  //-------------------------------------------------------
  // MEMBER FUNCTIONS -------------------------------------
  //-------------------------------------------------------
  virtual std::string get_name(){return name;};
  virtual std::string get_purename(){return purename;};
  virtual void measure(){};
  virtual void take(Measurement*){};
  virtual void take_dE (Measurement*){};
  virtual void take_nO (Measurement*){};
  virtual void take_daE(Measurement*){};
  virtual void take_naO(Measurement*){};
  virtual void devour(short,int,double,double,short,short){};
  virtual void add() {};
  virtual void add(double&,double,double) {};
  virtual void reset(){};
  virtual void reset(int){};
  virtual void sumup(){};
  virtual void setname(std::string base) {name = base + "_" + name;};

  virtual void screen(double norm) {screen(norm,false);};
  virtual void screen(double norm, bool stdcerr) {
    double mean,err;
    statistics(mean,err,norm);
    if (stdcerr)
      std::cerr << std::setw(20) << name << ": " << std::setw(20) << mean << "   " << std::setw(20) << err << std::endl;
    else
      std::cout << std::setw(20) << name << ": " << std::setw(20) << mean << "   " << std::setw(20) << err << std::endl;
  };

  void write(std::ofstream& datafile, double v) {datafile << v << std::endl;};
  void write(double N, double beta) {write(N*beta);};
  virtual void write(double norm){
    filename = name;
    filename.append(".dat");
    std::ofstream datafile(filename.c_str(),std::ios::app);
    double val;
    while (!empty()) {
      write(datafile,front()/norm);
      pop_front();
    }
  };

  void statistics(int &N, double& mean,double& err,double norm) {
    if (empty()) std::cerr << "No measurements have been defined or taken for " << name << std::endl;
    else {
      if (N>size()) N = size();
      mean = 0.;
      err = 0.;
      iter = end();
      for (int i=0;i<N;i++) {
	iter--;
	mean += *iter;
	err += (*iter)*(*iter);
      }
      
      mean /= N*norm;
      err /= N*norm*norm;
      err = sqrt((err-mean*mean)/(N-1.));
    }
  };
  void statistics(double& mean,double& err,double norm) {
    if (empty()) std::cerr << "No measurements have been defined or taken for " << name << std::endl;
    else {
      double N = (double) size();
      mean = 0.;
      err = 0.;
      
      iter = begin();
      while (iter != end()) {
	mean += *iter;
	err += (*iter)*(*iter);
	iter++;
      }
      
      mean /= N*norm;
      err /= N*norm*norm;
      err = sqrt((err-mean*mean)/(N-1.));
    }
  };
  void statistics(double& mean,double& err) {statistics(mean,err,1.);};
  //-------------------------------------------------------
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// DIAGONAL ENERGY %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class DiagEnergy : public Measurement {
 public:
  DiagEnergy(Spinsystem &sa):Measurement(sa,(std::string)"dE"){};
  ~DiagEnergy(){clear();};
  
  void measure() {
    if (done) done = false;
    else {
      push_back(0.);
      for (uint i=0;i<SA->bond.size();i++) back() += SA->bond[i]->diagenergy(SA->t0);
    }
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// DIAGONAL ENERGY ON ALPHABONDS %%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class DiagAlphaEnergy : public Measurement {
 public:
  DiagAlphaEnergy(Spinsystem &sa):Measurement(sa,(std::string)"daE"){};
  DiagAlphaEnergy(Spinsystem &sa,std::string name):Measurement(sa,name){};
  ~DiagAlphaEnergy(){clear();};
  
  void measure() {
    if (done) done = false;
    else {
      push_back(0.);
      for (uint i=0;i<SA->subbond.size();i++) if (SA->subbond[i]->ab) back() += SA->subbond[i]->diagenergy(SA->t0);
    }
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// NUMBER OF BREAKUPS/OPERATORS %%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class NumOp : public Measurement {
  bool otf;
 public:
  NumOp(Spinsystem &sa):Measurement(sa,(std::string)"nO"),otf(true){};
  NumOp(Spinsystem &sa,std::string name):Measurement(sa,name),otf(true){};
  NumOp(Spinsystem &sa,std::string name,bool o):Measurement(sa,name),otf(o){};
  ~NumOp(){clear();};
  
  void measure() {
    if (done) done = false;
    else {
      if (otf) push_back((double)SA->nO);
      else {
	push_back(0.);
	for (uint i=0;i<SA->bond.size();i++) back() += SA->bond[i]->numbonds(SA->t0);
      }
    }
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// NUMBER OF BREAKUPS/OPERATORS ON ALPHABONDS %%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class NumAlphaOp : public Measurement {
  bool otf;
 public:
  NumAlphaOp(Spinsystem &sa):Measurement(sa,(std::string)"naO"),otf(true){};
  NumAlphaOp(Spinsystem &sa,std::string name):Measurement(sa,name),otf(true){};
  NumAlphaOp(Spinsystem &sa,std::string name,bool o):Measurement(sa,name),otf(o){};
  ~NumAlphaOp(){clear();};
  
  void measure() {
    if (done) done = false;
    else {
      if (otf) push_back((double)SA->naO);
      else {
	push_back(0.);
	for (uint i=0;i<SA->subbond.size();i++) if (SA->subbond[i]->ab) back() += SA->subbond[i]->numbonds(SA->t0);
      }
    }
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// ENERGY ON ALPHA BONDS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class AlphaEnergy : public Measurement {
  std::vector<double> vec_aE,vec_aO;
  Measurement *daE,*naO;
  double aE,aO;

 public:
  AlphaEnergy(Spinsystem &sa):Measurement(sa,(std::string)"aE"),daE(NULL),naO(NULL){};
  ~AlphaEnergy(){clear();};

  void measure() {
    if (done) done = false;
    else {
      aE = 0.;
      aO = 0.;
      
      Subbond *sb;
      for (uint i=0;i<SA->subbond.size();i++) {
	sb = SA->subbond[i];
	if (sb->ab) {
	  vec_aE.push_back(sb->diagenergy(SA->t0));
	  vec_aO.push_back(sb->numbonds(SA->t0));
	}
      }
      
      for (uint i=0;i<vec_aE.size();i++) aE += vec_aE[i];
      for (uint i=0;i<vec_aO.size();i++) aO += vec_aO[i];
      
      push_back(aE-aO);
      
      if (daE != NULL) {
	daE->push_back(aE);
	daE->done = true;
      }
      if (naO != NULL) {
	naO->push_back(aO);
	naO->done = true;
      }
      
      vec_aE.clear();
      vec_aO.clear();
    }
  };
  
  void take_daE(Measurement* m){daE = m;};
  void take_naO(Measurement* m){naO = m;};
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// ENERGY %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Energy : public Measurement {
  std::vector<double> vec_udE,vec_adE,vec_uO,vec_aO,vec_aE;
  Measurement *dE,*nO,*daE,*naO,*aE;
  double udE,adE,uO,aO;

 public:
  Energy(Spinsystem &sa):Measurement(sa,(std::string)"E"),dE(NULL), nO(NULL),daE(NULL),naO(NULL),aE(NULL){};
  ~Energy(){clear();};

  void measure() {
    udE = 0.;
    adE = 0.;
    uO = 0.;
    aO = 0.;

    Subbond *sb;
    for (uint i=0;i<SA->subbond.size();i++) {
      sb = SA->subbond[i];
      if (sb->ab) {
	vec_adE.push_back(sb->diagenergy(SA->t0));
	vec_aO.push_back(sb->numbonds(SA->t0));
      }
      else {
	vec_udE.push_back(sb->diagenergy(SA->t0));
	vec_uO.push_back(sb->numbonds(SA->t0));
      }
    }

    for (uint i=0;i<vec_udE.size();i++) udE += vec_udE[i];
    for (uint i=0;i<vec_adE.size();i++) adE += vec_adE[i];
    for (uint i=0;i<vec_uO.size();i++) uO += vec_uO[i];
    for (uint i=0;i<vec_aO.size();i++) aO += vec_aO[i];
    
    push_back(udE+adE-uO-aO);
    
    if (aE != NULL) {
      aE->push_back(adE-aO);
      aE->done = true;
    }
    if (dE != NULL) {
      dE->push_back(udE+adE);
      dE->done = true;
    }
    if (nO != NULL) {
      nO->push_back(uO+aO);
      nO->done = true;
    }
    if (daE != NULL) {
      daE->push_back(adE);
      daE->done = true;
    }
    if (naO != NULL) {
      naO->push_back(aO);
      naO->done = true;
    }

    vec_udE.clear();
    vec_uO.clear();
    vec_adE.clear();
    vec_aO.clear();
  };

  void take_aE (Measurement* m){aE  = m;};
  void take_dE (Measurement* m){dE  = m;};
  void take_nO (Measurement* m){nO  = m;};
  void take_daE(Measurement* m){daE = m;};
  void take_naO(Measurement* m){naO = m;};
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// MAGNETIZATION %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Mag : public Measurement {
 public:
  Mag(Spinsystem &sa):Measurement(sa,(std::string)"m"){};
  Mag(Spinsystem &sa,std::string name):Measurement(sa,name){};
  ~Mag(){clear();};

  virtual void add() {push_back(res);};
  
  // at tau=0
  void measure(){
    res=0.; for(s=SA->spin.begin();s!=SA->spin.end();s++) res += (double)(*s)->state(); res /= 2.;
    add();
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// STAGGERED MAGNETIZATION %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Stmag : public Measurement {
 public:
  bool slice;        // slice or integrated meas?
  std::map<double,short>::iterator iter,before;

  Stmag(Spinsystem &sa):Measurement(sa,(std::string)"sl_stm"),slice(true){};
  Stmag(Spinsystem &sa,std::string name):Measurement(sa,name),slice(true){};
  Stmag(Spinsystem &sa, bool sl):Measurement(sa,(sl?(std::string)"sl_stm":(std::string)"in_stm")),slice(sl){};
  Stmag(Spinsystem &sa,std::string name,bool sl):Measurement(sa,name),slice(sl){};
  ~Stmag(){clear();};

  virtual void add() {push_back(res);};
  virtual void add(double& r,double term,double time) {r += term*time;};

  // at tau=0
  void measure(){(slice)?measure_sl():measure_in();};

  void measure_sl() {
    res = 0.; 
    for (s=SA->spin.begin();s!=SA->spin.end();s++) res += ((*s)->stag)?(double)-(*s)->state():(double)(*s)->state();
    res /= 2.;
    add();
  };

  void measure_in(){
    res = 0.;
    SA->get_intervals();

    iter = SA->intervals.begin();
    before = iter;

    double time(iter->first - SA->t0);
    double val(0.);
    for (s=SA->spin.begin();s!=SA->spin.end();s++) {
      if ((*s)->stag) val -= (double)(*s)->state();
      else val += (double)(*s)->state();
    }
    val /= 2.;

    add(res,val,time);
    iter++;

    int bd;
    for (;iter!=SA->intervals.end();iter++) {
      time = iter->first - before->first;
      bd = before->second;
      if (bd > 0) {
	if (SA->bond[bd-1]->Sp[0]->stag) val += 1.;
	else val -= 1.;
	if (SA->bond[bd-1]->Sp[1]->stag) val -= 1.;
	else val += 1.;
      }
      else if (bd < 0) {
	if (SA->bond[-bd-1]->Sp[0]->stag) val -= 1.;
	else val += 1.;
	if (SA->bond[-bd-1]->Sp[1]->stag) val += 1.;
	else val -= 1.;
      }

      add(res,val,time);
      before = iter;
    }

    push_back(res);
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// ABSOLUTE STAGGERED MAGNETIZATION %%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class AbsStmag : public Stmag {
 public:
  AbsStmag(Spinsystem &sa):Stmag(sa,(std::string)"sl_stm0"){};
  AbsStmag(Spinsystem &sa, bool sl):Stmag(sa,(sl?(std::string)"sl_stm0":(std::string)"in_stm0"),sl){};
  ~AbsStmag(){clear();};

  void add() {push_back(fabs(res));};
  void add(double& r,double term,double time) {r += fabs(term)*time;};
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SQUARED STAGGERED MAGNETIZATION %%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class SqStmag : public Stmag {
 public:
  SqStmag(Spinsystem &sa):Stmag(sa,(std::string)"sl_stm2"){};
  SqStmag(Spinsystem &sa, bool sl):Stmag(sa,(sl?(std::string)"sl_stm2":(std::string)"in_stm2"),sl){};
  ~SqStmag(){clear();};

  void add() {push_back(res*res);};
  void add(double& r,double term,double time) {r += term*term*time;};
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// QUARTIC STAGGERED MAGNETIZATION %%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class QuStmag : public Stmag {
 public:
  QuStmag(Spinsystem &sa):Stmag(sa,(std::string)"sl_stm4"){};
  QuStmag(Spinsystem &sa, bool sl):Stmag(sa,(sl?(std::string)"sl_stm4":(std::string)"in_stm4"),sl){};
  ~QuStmag(){clear();};

  void add() {push_back(res*res*res*res);};
  void add(double& r,double term,double time) {r += term*term*term*term*time;};
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// TWIST ORDER PARAMETER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Twist : public Measurement {
 public:
  bool slice;        // slice or integrated meas?
  std::map<double,short>::iterator iter,before;
  Measurement *taken1,*taken2;
  double res1,res2;

  Twist(Spinsystem &sa):Measurement(sa,(std::string)"sl_tw"),slice(true),taken1(NULL),taken2(NULL),res2(0.){};
  Twist(Spinsystem &sa, std::string name):Measurement(sa,name),slice(true),taken1(NULL),taken2(NULL),res2(0.){};
  Twist(Spinsystem &sa, bool sl):
    Measurement(sa,(sl?(std::string)"sl_tw":(std::string)"in_tw")),slice(sl),taken1(NULL),taken2(NULL),res2(0.){};
  Twist(Spinsystem &sa, std::string name, bool sl):Measurement(sa,name),slice(sl),taken1(NULL),taken2(NULL),res2(0.){};
  ~Twist(){clear();};

  // at tau=0
  void measure(){
    if (slice) measure_sl();
    else measure_in();
  };

  virtual void add() {push_back(res);};
  virtual void add(double& r,double term,double time) {r += term*time;};

  void measure_sl(){
    res = 0.;
    if (SA->Ly == 1) {
      for (s=SA->spin.begin();s!=SA->spin.end();s++) res += (double)((*s)->state() * ((*s)->col+1));
      res = cos((double)TWOPI/(double)SA->Lx*res/2.);
      add();
    }
  };

  void measure_in(){
    if (done) done = false;
    else {
      res = 0.;
      res1 = 0.;
      res2 = 0.;
      double term;
      SA->get_intervals();
      
      iter = SA->intervals.begin();
      before = iter;
      
      double time(iter->first - SA->t0);
      double val(0.);
      for (s=SA->spin.begin();s!=SA->spin.end();s++) val += (double)((*s)->state() * ((*s)->col+1));
      val /= 2.;

      term = cos((double)TWOPI/(double)SA->Lx*val);
      add(res,term,time);
      if (taken1 != NULL) taken2->add(res2,term,time);
      if (taken2 != NULL) taken2->add(res2,term,time);
      iter++;
      
      int bd;
      for (;iter!=SA->intervals.end();iter++) {
	time = iter->first - before->first;
	bd = before->second;
	if (bd > 0) {
	  val -= (double)(SA->bond[bd-1]->Sp[0]->col+1);
	  val += (double)(SA->bond[bd-1]->Sp[1]->col+1);
	}
	else if (bd < 0) {
	  val += (double)(SA->bond[-bd-1]->Sp[0]->col+1);
	  val -= (double)(SA->bond[-bd-1]->Sp[1]->col+1);
	}
	
	term = cos((double)TWOPI/(double)SA->Lx*val);
	add(res,term,time);
	if (taken1 != NULL) taken1->add(res1,term,time);
	if (taken2 != NULL) taken2->add(res2,term,time);
	before = iter;
      }
      
      push_back(res);
      if (taken1 != NULL) {
	taken1->push_back(res1);
	taken1->done = true;
      }
      if (taken2 != NULL) {
	taken2->push_back(res2);
	taken2->done = true;
      }
    }
  };

  void take(Measurement* m){
    if (taken1 == NULL) taken1 = m;
    else if (taken2 == NULL) taken2 = m;
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// ABS TWIST ORDER PARAMETER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class AbsTwist : public Twist {
 public:
  AbsTwist(Spinsystem &sa):Twist(sa,(std::string)"sl_tw0"){};
  AbsTwist(Spinsystem &sa,bool sl):Twist(sa,(sl)?(std::string)"sl_tw0":(std::string)"in_tw0",sl){};
  ~AbsTwist(){clear();};

  void add() {push_back(fabs(res));};
  void add(double& r,double term,double time) {r += fabs(term)*time;};
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SQUARED TWIST ORDER PARAMETER %%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class SqTwist : public Twist {
 public:
  SqTwist(Spinsystem &sa):Twist(sa,(std::string)"in_tw2"){};
  SqTwist(Spinsystem &sa,bool sl):Twist(sa,(sl)?(std::string)"sl_tw2":(std::string)"in_tw2",sl){};
  ~SqTwist(){clear();};

  void add() {push_back(res*res);};
  void add(double& r,double term,double time) {r += term*term*time;};
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// QUARTIC TWIST ORDER PARAMETER %%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class QuTwist : public Twist {
 public:
  QuTwist(Spinsystem &sa):Twist(sa,(std::string)"in_tw4"){};
  QuTwist(Spinsystem &sa,bool sl):Twist(sa,(sl)?(std::string)"sl_tw4":(std::string)"in_tw4",sl){};
  ~QuTwist(){clear();};

  void add() {push_back(res*res*res*res);};
  void add(double& r,double term,double time) {r += term*term*term*term*time;};
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// PAIR ANTI CORRELATION %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Pair : public Measurement {
  const short j;

 public:
  Spin *s0,*s1;

  Pair(Spinsystem &sa):Measurement(sa,(std::string)"pair"),j(0),s0(sa.spin[2*j]),s1(sa.spin[2*j+1]) {};
  Pair(Spinsystem &sa, short jj):Measurement(sa,(std::string)"pair"),j(jj),s0(sa.spin[2*j]),s1(sa.spin[2*j+1]) {};
  ~Pair(){clear();};

  // at tau=0
  void measure(){push_back(-1.*(double)(s0->state()*s1->state()));};
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// AVERAGED PAIR ANTI CORRELATION %%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class AvgPair : public Measurement {
 public:
  short sres;
  Spin *s0,*s1;

  AvgPair(Spinsystem &sa):Measurement(sa,(std::string)"av_pair"){};
  ~AvgPair(){clear();};

  // at tau=0
  void measure(){
    sres = 0;
    Spin *s0,*s1;
    for (short i=0;i<SA->spin.size()/2;i++) {
      s0 = SA->spin[2*i];
      s1 = SA->spin[2*i+1];
      sres -= s0->state()*s1->state();
    }

    push_back((double)sres);
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// HELPING CLASS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class HelpMatrix {
  double E,S,cosw,sinw,term;
  
 public:
  double kx,ky,w;
  std::vector<vec_d> cxcy_m_sxsy;
  std::vector<vec_d> sxcy_p_cxsy;

  HelpMatrix(Spinsystem &sa, double _kx, double _ky, double _w):kx(_kx),ky(_ky),w(_w){

    for (uint i=0;i<sa.Lx;i++) {
      cxcy_m_sxsy.push_back(vec_d());
      sxcy_p_cxsy.push_back(vec_d());
      for (uint j=0;j<sa.Ly;j++) {
	cxcy_m_sxsy[i].push_back(cos(kx*(double)i)*cos(ky*(double)j) - sin(kx*(double)i)*sin(ky*(double)j));
	sxcy_p_cxsy[i].push_back(sin(kx*(double)i)*cos(ky*(double)j) + cos(kx*(double)i)*sin(ky*(double)j));
      }
    }
    
  };
  ~HelpMatrix(){};

  void devour(double& RE, double& IM, int type, short _state,double _tS,double _tE,short _rx,short _ry) {
    switch (type) {
    case 1:
      E = w*_tE;
      S = w*_tS;
      cosw = cos(E)-cos(S);
      sinw = sin(E)-sin(S);
      term = (double)_state/w;
      E = cxcy_m_sxsy[_rx][_ry];
      S = sxcy_p_cxsy[_rx][_ry]; 
      RE += (E*sinw - S*cosw)*term;
      IM += (E*cosw + S*sinw)*term;
      break;
    case 2:
      term = (double)_state*(_tE-_tS);
      RE += term*cxcy_m_sxsy[_rx][_ry];
      IM += term*sxcy_p_cxsy[_rx][_ry];
      break;
    case 3:
      E = w*_tE;
      S = w*_tS;
      cosw = cos(E)-cos(S);
      sinw = sin(E)-sin(S);
      term = (double)_state/w;
      RE += term*sinw;
      IM += term*cosw;
      break;
    }
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// NUMBER OF LOOPS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class NumLoops : public Measurement {
  int loopID;
 public:
  NumLoops(Spinsystem &sa):Measurement(sa,(std::string)"nL"),loopID(-1){};
  NumLoops(Spinsystem &sa,std::string name):Measurement(sa,name),loopID(-1){};
  NumLoops(Spinsystem &sa,std::string name,int lid):Measurement(sa,name),loopID(lid){};
  ~NumLoops(){clear();};

  void measure() {
    if (loopID == -1) push_back((double)SA->loop.label);
    else push_back((SA->loop.label==loopID)?1.:0.);
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// AVERAGE LOOP SIZE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class AvgLoopSize : public Measurement {
 public:
  AvgLoopSize(Spinsystem &sa):Measurement(sa,(std::string)"als"){};
  AvgLoopSize(Spinsystem &sa,std::string name):Measurement(sa,name){};
  ~AvgLoopSize(){clear();};

  void measure() {push_back(1./(double)SA->loop.label);};
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SUM OF SQUARED LOOP SIZES %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class SqLoopSize : public Measurement {
  double tS;
  Subspin::iterator iter;

 public:
  std::vector<double> sizes;
  std::vector<Measurement*> taken;

  SqLoopSize(Spinsystem &sa):Measurement(sa,(std::string)"imp_stsuspm"){};
  SqLoopSize(Spinsystem &sa,std::string name):Measurement(sa,name){};
  ~SqLoopSize(){clear();};

  virtual void sumup() {
    for (int i=0;i<SA->loop.label;i++) res += sizes[i]*sizes[i];
    push_back(res);
  };

  virtual void devour(short _state,int label,double _tS,double _tE,short _rx,short _ry) {
    sizes[label-1] += (_tE-_tS);
  };

  void reset(int l) {
    sizes.clear(); res = 0.;
    for (int i=0;i<l;i++) sizes.push_back(0.);
  };

  virtual void measure() {
    reset(SA->loop.label);

    for (uint t=0;t<taken.size();t++) taken[t]->reset(SA->loop.label);
    
    for (uint i=0;i<SA->spin.size();i++) {
      for (uint j=0;j<SA->spin[i]->size();j++) {
	tS = SA->t0;
	for (iter =(*SA->spin[i])[j]->begin();iter!=(*SA->spin[i])[j]->end();iter++) {
	  devour(((SA->spin[i]->stag)?-(*iter)->state:(*iter)->state),
		 (*iter)->label,tS,(*iter)->t,SA->spin[i]->col,SA->spin[i]->row);
	  for(uint t=0;t<taken.size();t++) taken[t]->devour(((SA->spin[i]->stag)?-(*iter)->state:(*iter)->state),
							    (*iter)->label,tS,(*iter)->t,SA->spin[i]->col,SA->spin[i]->row);
	  tS = (*iter)->t;
	}
      }
    }
    
    sumup();
    
    for (uint t=0;t<taken.size();t++) {
      taken[t]->sumup();
      taken[t]->done = true;
    }
  };

  void take(Measurement* m){taken.push_back(m);};
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SUM OF SQUARED LOOP WEIGHTS     %%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class SqLoopWeight : public SqLoopSize {
 public:
  SqLoopWeight(Spinsystem &sa):SqLoopSize(sa,(std::string)"imp_stsus"){};
  SqLoopWeight(Spinsystem &sa,std::string name):SqLoopSize(sa,name){};
  ~SqLoopWeight(){clear();};

  void devour(short _state,int label,double _tS,double _tE,short _rx,short _ry) {
    sizes[label-1] += (double)_state*(_tE-_tS);
  };
};
// for compatibility with older versions:
typedef SqLoopWeight SqLoopSizePM;
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SUM OF SQUARED CLUSTER SIZES    %%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class SqClusterSize : public SqLoopSize {
 public:
  SqClusterSize(Spinsystem &sa):SqLoopSize(sa,(std::string)"imp_stsus"){};
  SqClusterSize(Spinsystem &sa,std::string name):SqLoopSize(sa,name){};
  ~SqClusterSize(){clear();};

  void sumup() {
    for (int i=0;i<SA->glue.glue.size();i++) {
      if (SA->glue.glue[i] != i+1) {
	sizes[SA->glue.glue[i]-1] += sizes[i];
	sizes[i] = 0.;
      }
    }
    for (int i=0;i<SA->glue.glue.size();i++) res += sizes[i]*sizes[i];
    push_back(res);
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SUM OF SQUARED CLUSTER WEIGHTS  %%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class SqClusterWeight : public SqClusterSize {
 public:
  SqClusterWeight(Spinsystem &sa):SqClusterSize(sa,(std::string)"imp_stsus"){};
    SqClusterWeight(Spinsystem &sa,std::string name):SqClusterSize(sa,name){};
  ~SqClusterWeight(){clear();};

  void devour(short _state,int label,double _tS,double _tE,short _rx,short _ry) {
    sizes[label-1] += (double)_state*(_tE-_tS);
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SUM OF SQUARED GENERAL LOOP SIZES %%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class GenSqLoopSize : public Measurement {
  double kx,ky,w,tS;
  Subspin::iterator iter;
  const double zero;

 public:
  int type;
  HelpMatrix hM;
  std::vector<double> RE_sizes,IM_sizes;

  GenSqLoopSize(Spinsystem &sa, std::string name,double _kx, double _ky, double _w):
    Measurement(sa,name),kx(_kx),ky(_ky),w(_w),zero(0.),type(1),hM(sa,_kx,_ky,_w) {
    if (w == zero) type = 2;
    else if (kx == zero && ky == zero) type = 3;
  };
  ~GenSqLoopSize(){clear();};

  virtual void sumup() {
    for (int i=0;i<RE_sizes.size();i++) res += (RE_sizes[i]*RE_sizes[i]+IM_sizes[i]*IM_sizes[i]);
    push_back(res);
  };

  virtual void devour(short _state,int label,double _tS,double _tE,short _rx,short _ry) {
    hM.devour(RE_sizes[label-1],IM_sizes[label-1],type,1,_tS,_tE,_rx,_ry);
  };

  void reset(int l) {
    res = 0.;
    RE_sizes.clear();
    IM_sizes.clear();
    for (int i=0;i<l;i++) {
      RE_sizes.push_back(0.);
      IM_sizes.push_back(0.);
    }
  };

  void measure() {
    if (done) done = false;
    else {
      reset(SA->loop.label);
      for (uint i=0;i<SA->spin.size();i++) {
	for (uint j=0;j<SA->spin[i]->size();j++) {
	  tS = SA->t0;
	  for (iter =(*SA->spin[i])[j]->begin();iter!=(*SA->spin[i])[j]->end();iter++) {
	    devour(((SA->spin[i]->stag)?-(*iter)->state:(*iter)->state),
		   (*iter)->label,
		   tS,
		   (*iter)->t,
		   SA->spin[i]->col,
		   SA->spin[i]->row);
	    tS = (*iter)->t;
	  }
	}
      }
      sumup();
    }
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SUM OF SQUARED GENERAL LOOP WEIGHTS %%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class GenSqLoopWeight : public GenSqLoopSize {
 public:
  GenSqLoopWeight(Spinsystem &sa, std::string name,double _kx, double _ky, double _w):GenSqLoopSize(sa,name,_kx,_ky,_w){};
  ~GenSqLoopWeight(){clear();};

  void devour(short _state,int label,double _tS,double _tE,short _rx,short _ry) {
    hM.devour(RE_sizes[label-1],IM_sizes[label-1],type,_state,_tS,_tE,_rx,_ry);
  };
};
// for compatibility with older versions:
typedef GenSqLoopWeight GenSqLoopSizePM;
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SUM OF SQUARED GENERAL CLUSTER SIZES %%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class GenSqClusterSize : public GenSqLoopSize {
 public:
  GenSqClusterSize(Spinsystem &sa, std::string name,double _kx, double _ky, double _w):GenSqLoopSize(sa,name,_kx,_ky,_w){};
  ~GenSqClusterSize(){clear();};

  void sumup() {
    for (int i=0;i<SA->glue.glue.size();i++) {
      if (SA->glue.glue[i] != i+1) {
	RE_sizes[SA->glue.glue[i]-1] += RE_sizes[i];
	RE_sizes[i] = 0.;
	IM_sizes[SA->glue.glue[i]-1] += IM_sizes[i];
	IM_sizes[i] = 0.;
      }
    }
    for (int i=0;i<SA->glue.glue.size();i++) res += (RE_sizes[i]*RE_sizes[i]+IM_sizes[i]*IM_sizes[i]);
    push_back(res);
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SUM OF SQUARED GENERAL CLUSTER WEIGHTS %%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class GenSqClusterWeight : public GenSqClusterSize {
 public:
  GenSqClusterWeight(Spinsystem &sa, std::string name,double _kx, double _ky, double _w):GenSqClusterSize(sa,name,_kx,_ky,_w){};
  ~GenSqClusterWeight(){clear();};

  void devour(short _state,int label,double _tS,double _tE,short _rx,short _ry) {
    hM.devour(RE_sizes[label-1],IM_sizes[label-1],type,_state,_tS,_tE,_rx,_ry);
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SUM OF SQUARED TEMPORAL WINDING NUMBER OF LOOPS %%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class TempWindNum2 : public Measurement {
 public:
  TempWindNum2(Spinsystem &sa):Measurement(sa,(std::string)"imp_twN2"){};
  ~TempWindNum2(){clear();};
  
  void measure() {
    vec_i wN(SA->loop.label+1);
    std::map<short,Subspin*>::iterator sb;
    sb = SA->subspin.begin();
    for (;sb!=SA->subspin.end();sb++) 
      wN[sb->second->getLabelAt(0.)] += sb->second->state();
    int res(0);
    for (uint i=0;i<wN.size();i++) res += wN[i]*wN[i];
    push_back((double)res);
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SUM OF SQUARED SPATIAL WINDING NUMBER OF LOOPS %%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class SpatialWindNum2 : public Measurement {
 public:
  SpatialWindNum2(Spinsystem &sa):Measurement(sa,(std::string)"imp_swN2"){};
  ~SpatialWindNum2(){clear();};
  
  void measure() {
    vec_i wN(SA->loop.label+1);
    SA->bond[0]->get_spatialWindNum(wN);
    int res(0);
    for (uint i=0;i<wN.size();i++) res += wN[i]*wN[i];
    push_back((double)res);
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



#endif // MEASUREMENTS_H ///:~
