/**
 * \file simulation.h
 *
 * \brief Contains a wrapping class for simulation.
 * Defined for chain models only.
 *
 */

#ifndef SIMULATION_H
#define SIMULATION_H

#include "measurement_producer.h"




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SIMULATION %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Simulation {

  bool com(std::string s1, std::string s2) {return (s1.compare(s2)==0);};

 public:

  //-------------------------------------------------------
  // DIRECT PARAMETERS ------------------------------------
  //-------------------------------------------------------
  std::string sModel;	/**< \brief defines the model, a chain here! */
  std::string filebase;	/**< \brief to be used as base for filenames */
  uint Lx;		/**< \brief length of the chain */
  uint Nmeas;		/**< \brief number of measurements */
  double beta;		/**< \brief inverse temperature = length of imaginary time */
  double alpha;		/**< \brief bond alternation */
  double Delta;		/**< \brief XXZ exchange anisotropy */
  //-------------------------------------------------------
  // PARAMETERS FROM FILE ---------------------------------
  //-------------------------------------------------------
  std::string sRng;	/**< \brief name of RNG */
  uint Ly;		/**< \brief defaults to 1 */
  uint Ntherm;		/**< \brief number of updates for thermalization */
  uint num;		/**< \brief =Nmeas+Ntherm */
  uint writeout;	/**< \brief increment of updates at which data is written to disk */
  bool screen;		/**< \brief print mesurement results to screen? (0=no, 1=yes) */
  bool write;		/**< \brief write mesurements to file? (0=no, 1=yes) */
  bool writestate;	/**< \brief write state to file? (0=no, 1=yes) */
  bool sow;		/**< \brief generate seeds for RNG (0=no, 1=yes) */
  bool logging;		/**< \brief show logging information (0=no, 1=yes) */
  //-------------------------------------------------------
  
  Spinsystem *SA;
  std::vector<Measurement*> uni;/**< \brief container for unimproved estimators */
  std::vector<Measurement*> imp;/**< \brief container for improved estimators */
  std::vector<double> norm_uni;	/**< \brief container for the norm of unimproved estimators */
  std::vector<double> norm_imp;	/**< \brief container for the norm of improved estimators */
  uint n_uni;			/**< \brief number of unimproved estimators */
  uint n_imp;			/**< \brief number of improved estimators */

  Simulation(std::string _sModel,
	     uint        _Lx,
	     double      _beta,
	     double      _alpha,
	     double      _Delta,
	     uint        _Nmeas,
	     std::string _filebase):
    sModel              (_sModel),
    filebase            (_filebase),
    Lx                  (_Lx),
    Nmeas               (_Nmeas),
    beta                (_beta),
    alpha               (_alpha),
    Delta               (_Delta),
    sRng                ("RAND"),
    Ly                  (1),
    Ntherm              (10000),
    num                 (Ntherm+Nmeas),
    writeout            (1000),
    screen              (true),
    write               (false),
    writestate          (false),
    sow                 (true),
    logging             (true),
    n_imp               (0),
    n_uni               (0)
  {};
  ~Simulation(){delete SA;};

  void read_parameter(std::string file) {
    std::ifstream in(file.c_str());
    std::string s1,s2;
    if (!in.is_open()) {
      showlog("Could not open parameter file: " << file);
      showlog("Using default parameters instead.");
    }
    else {
      showlog("Parameter file:" << file);
      while (!in.eof()) {
	in >> s1;
	if (!s1.empty() && *s1.begin() != '#') {
	  in >> s2;
	  if      (com(s1,"sRng"))		sRng		= s2;
	  else if (com(s1,"Ly"))		Ly		= atoi(s2.c_str());
	  else if (com(s1,"Ntherm"))		Ntherm		= atoi(s2.c_str());
	  else if (com(s1,"screen"))		screen		= (s2.compare("true")==0);
	  else if (com(s1,"write"))		write		= (s2.compare("true")==0);
	  else if (com(s1,"writestate"))	writestate	= (s2.compare("true")==0);
	  else if (com(s1,"sow"))		sow		= (s2.compare("true")==0);
	  else if (com(s1,"logging"))		logging		= (s2.compare("true")==0);
	  else if (com(s1,"writeout"))		writeout	= atoi(s2.c_str());
	}
	else in >> s1;
      }
    }
    num = Nmeas+Ntherm;
    in.close();

    if (logging) {
      showlog("Parameters:");
      showlog("sModel   = " << sModel);
      showlog("filebase = " << filebase);
      showlog("Lx       = " << Lx);
      showlog("Ly       = " << Ly);
      showlog("beta     = " << beta);
      showlog("alpha    = " << alpha);
      showlog("Delta    = " << Delta);
      showlog("sRng     = " << sRng);
      showlog("Ntherm   = " << Ntherm);
      showlog("Nmeas    = " << Nmeas);
      showlog("num      = " << num);
      showlog("--------------------");
      showlog("screen       : "	<< ((screen)?		"YES":"NO"));
      showlog("write        : "	<< ((write)?		"YES":"NO"));
      showlog("writestate   : "	<< ((writestate)?	"YES":"NO"));
      showlog("sow          : "	<< ((sow)?		"YES":"NO"));
      showlog("logging      : "	<< ((logging)?		"YES":"NO"));
    }
  };

  void init_spinsystem() {init_spinsystem((unsigned long)0);}
  void init_spinsystem(unsigned long seed) {
    if (seed == 0) exp_dist.load(sRng,sow); // exp_dist contains the RNG, sRng is the name
    else exp_dist.load(sRng,seed);

    // always write the state of the RNG to screen
    std::vector<unsigned long> state;
    exp_dist.rng->get_state(state);
    showlog("State of " << exp_dist.rng->get_name() << ":");
    for (int i=0;i<state.size();i++) showlog(state[i]);

    Basecell bc;			
    MODEL model(bc.translate(sModel));	// the basecell must generate the model
    bc.generate(model,alpha,Delta);	// and be given to the spinssystem
    SA = new Spinsystem(Lx/bc.Lx,Ly/bc.Ly,bc,(double)0.,beta);
  };

  uint init_uni_estimator(std::string name_of_measurement) {
    MeasurementProducer mp(*SA);
    uni.push_back(mp.get_measurement(name_of_measurement));
    norm_uni.push_back(mp.get_norm(name_of_measurement));
    n_uni++;
  }
  uint init_uni_estimators(std::string name_of_measurement) {
    MeasurementProducer mp(*SA);
    mp.read_measurements(name_of_measurement+".uni",uni,norm_uni,"dnaO"); // dnaO MUST not be read in here!!!
    if (logging) for (uint i=0;i<uni.size();i++) showlog(uni[i]->get_name() << " loaded with norm " << norm_uni[i]);
    for (uint i=0;i<uni.size();i++) uni[i]->setname(filebase);
    return n_uni = uni.size();
  };

  uint init_imp_estimator(std::string name_of_measurement) {
    MeasurementProducer mp(*SA);
    imp.push_back(mp.get_measurement(name_of_measurement));
    norm_imp.push_back(mp.get_norm(name_of_measurement));
    n_imp++;
  }
  uint init_imp_estimators(std::string name_of_measurement) {
    MeasurementProducer mp(*SA);
    mp.read_measurements(name_of_measurement+".imp",imp,norm_imp);
    if (logging) for (uint i=0;i<imp.size();i++) showlog(imp[i]->get_name() << " loaded with norm " << norm_imp[i]);
    for (uint i=0;i<imp.size();i++) imp[i]->setname(filebase);
    return n_imp = imp.size();
  };

  bool check_measurements() {return (uni.size()+imp.size()>0);};
  
  /**
   * \brief For debugging.
   */
  void decorate_help(int i, int I) {
    std::cout << std::setprecision(10);
    Displayer display;
    Checker check;
    int res(0);

    show("swapconnect started");
    SA->forEachSpin(SA->swapconnect);
    show("swapconnect finished");


    if (I>=2) SA->forEachSpin(display);
    show("updateOP started");
    SA->forEachBond(SA->updateOP);
    show("updateOP finished");

    check.check = "zerolinks";
    show(check.check << " started");
    res = 0;
    res = SA->forEachSubspin(check);

    if (res != 0) {
      show("checking " << check.check << ": " << res);
      if (I>=2) SA->forEachSpin(display);
      exit(0);
    }
    show(check.check << " finished");

    check.check = "inot";
    show(check.check << " started");
    res = 0;
    res = SA->forEachSubspin(check);
    if (res != 0) {
      show("checking " << check.check << ": " << res);
    }
    show(check.check << " finished");

    if (I>=3) {
      show("helpspread started");
      SA->helpspread.naO = 0;
      SA->nO  += SA->forEachBond(SA->helpspread);
      show("helpspread finished");
    } else {
      show("spread started");
      SA->spread.naO = 0;
      SA->nO  += SA->forEachBond(SA->spread);
      show("spread finished");
    }
      
    SA->naO += SA->spread.naO;
    SA->loop.prepared = false;

    show("loop 1 started");
    SA->forEachSpin(SA->loop);
    SA->loop.prepared = true;
    show("loop 2 started");
    SA->forEachSpin(SA->loop);
    show("loop finished");

    if (I>=2) SA->forEachSpin(display);

    SA->glue.init(SA->loop.label);
    SA->forEachBond(SA->glue);
    SA->glue.gluedown();
  };

  void measure_imp() {for (int j=0;j<n_imp;j++) imp[j]->measure();};
  void measure_uni() {for (int j=0;j<n_uni;j++) uni[j]->measure();};

  void write_out() {
    for (int j=0;j<n_imp;j++) imp[j]->write(norm_imp[j]);
    for (int j=0;j<n_uni;j++) uni[j]->write(norm_uni[j]);
    if (writestate) SA->write_state();
  };

  void simulate() {
    for (uint i=0;i<num;i++) {
      SA->decorate();							// DECORATION
      if (i >= Ntherm) measure_imp();					// MEASUREMENT IMPROVED ESTIMATORS
      SA->update();							// UPDATE
      if (i >= Ntherm) measure_uni();					// MEASUREMENT STANDARD ESTIMATORS
      if (write && i>=Ntherm && i%writeout==(writeout-1)) write_out();	// WRITE OUT 
    }
  };

  void simulate_help() {simulate_help(-1);};
  void simulate_help(int I) {
    Checker check;
    int res;
    myFile file;
    for (uint i=0;i<num;i++) {
      if (I>=0) showlog(i);
      if (I>=1) show("run " << i);
      
      check.check = "boundary";
      res = SA->forEachSpin(check);
      if (res != 0) show("checking " << check.check << ": " << res);
      
      if (I>=1) {
	show("decoration started");
	decorate_help(0,I);
	show("decoration finished");
      } else SA->decorate();
      if (i >= Ntherm) {measure_imp();}
      if (I>=1) show("update started");
      SA->update();
      if (I>=1) show("update finished");
      if (i >= Ntherm) measure_uni();
      if (write && i>=Ntherm && i%writeout==(writeout-1)) write_out();
    }
  };

  void finish() {
    if (screen) {
      for (int i=0;i<n_uni;i++) {
	showl(sModel << " " << Lx << " " << beta << " " << alpha << " " << Delta << "\t");
	uni[i]->screen(norm_uni[i]);
      }
      for (int i=0;i<n_imp;i++) {
	showl(sModel << " " << Lx << " " << beta << " " << alpha << " " << Delta << "\t");
	imp[i]->screen(norm_imp[i]);
      }
    }

    if (write) {
      write_out();
      if (writestate) SA->write_state();
    }
  };
  
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




#endif // SIMULATION_H ///:~
